#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include "slog_impl.h"                  /*I  "slog_impl.h"  I*/

/*@C
    SLOG_OpenOutputStream - Returns Output SLOG_STREAM associcated with
                            the given filename.

  Unmodified Input Variables :
. filename - constant char pointer to the filename.

  Modified Output Variables :
. returned value - pointer to the newly allocated SLOG_STREAM.  It is
                   NULL when the routine encounters error.

  Include File Needed :
  slog.h

.N SLOG_RETURN_STATUS
@*/
SLOG_STREAM *SLOG_OpenOutputStream( const char *filename )
{
    SLOG_STREAM *slog;


    SLOG_InitGlobalData();

    /*  memory for slog is allocated by SLOG_OpenStream()  */
    if ( ( slog = SLOG_OpenStream( filename, "wb" ) ) == NULL )
        return NULL;

    if ( SLOG_WriteHeader( slog ) != SLOG_SUCCESS )
        return NULL;
    if ( SLOG_WritePreview( slog ) != SLOG_SUCCESS )
        return NULL;
    if ( SLOG_WriteProfile( slog, SLOG_PROFILENAME ) != SLOG_SUCCESS )
        return NULL;

    return slog;
}



/*@C
    SLOG_Irec_ToOutputStream - store the given interval record in the frame
                               buffer and write it out to the disk when
                               the frame buffer is full.

   Modified Input Variables :
. slog - pointer to SLOG_STREAM where interval record will be stored.

   Unmodified Input Variables :
. irec - pointer to constant interval record with the fields set by 
         the user before this routine is called.

   Modified Output Variables :
. returned value - integer return status.

   Usage Notes on this subroutine :
   The routine provides all the core management of the implementation of
   SLOG_STREAM.  Be sure to call SLOG_Irec_SetMinRec() at least once for
   the interval record to be appended to the SLOG_STREAM.

  Include File Needed :
  slog.h

.N SLOG_RETURN_STATUS
@*/
int SLOG_Irec_ToOutputStream(       SLOG_STREAM      *slog,
                              const SLOG_intvlrec_t  *irec )
{
    SLOG_uint32     old_buf_Nbytes_in_file;
    SLOG_uint32     new_buf_Nbytes_in_file;

    SLOG_uint32     Nbytes_written;
    int             ierr;

#if defined( DEBUG )
    long slogfile_loc_cur;
#endif

    if ( slog->flagfirstentry == 1 ) {
        /*  Make sure this block is executed once for each SLOG_STREAM  */
        slog->flagfirstentry = 0;  
        ierr = SLOG_STM_CreateBbufs( slog );
        if ( ierr != SLOG_SUCCESS )
            return ierr;

        /*  Update the content of SLOG_hdr_t to both memory and disk  */
        slog->hdr->fptr2framedata = ( SLOG_fptr ) ftell( slog->fd );
        ierr = SLOG_WriteUpdatedHeader( slog );
        if ( ierr != SLOG_SUCCESS ) {
            fprintf( errfile,  __FILE__":SLOG_Irec_ToOutputStream() - "
                                       "SLOG_WriteUpdatedHeader() fails\n" );
            fflush( errfile );
            return ierr;
        }

        /*  Write the content of SLOG Preview to the disk  */

        /*
           SLOG_STM_Init() can be called either __BEFORE__ or _AFTER_ 
           SLOG_WriteInitFrameDir().  SLOG_STM_Init() initializes the 
           Frame_Dir components and various Bbufs in the SLOG_STREAM.  
           So SLOG_WriteInitFrameDir() and SLOG_STM_Init() are independent
           from one another.
        */
        SLOG_STM_Init( slog );
        ierr = SLOG_WriteInitFrameDir( slog );
        if ( ierr != SLOG_SUCCESS )
            return ierr;
    }   /*  Endof { if ( slog->flagfirstentry == 1 ) }  */

  /*  Check if the input interval record is empty  */
    if ( SLOG_Irec_IsEmpty( irec ) ) {
        fprintf( stdout, __FILE__":SLOG_Irec_ToOutputStream() - \n" );
        fprintf( stdout, "      ""Input Interval Record is empty\n" );
        fprintf( stdout, "      ""irec = " );
        SLOG_Irec_Print( irec, stdout ); fprintf( stdout, "\n" );
        fflush( stdout );
        return SLOG_FAIL;
    }

#if defined( DEBUG )
    if ( SLOG_Irec_IsConsistent( irec ) != SLOG_SUCCESS ) {
        fprintf( stdout, __FILE__":SLOG_Irec_ToOutputStream() - \n" );
        fprintf( stdout, "      ""Input Interval Record is InConsistent\n" );
        fprintf( stdout, "      ""irec = " );
        SLOG_Irec_Print( irec, stdout ); fprintf( stdout, "\n" );
        fflush( stdout );
        return SLOG_FAIL;
    }
#endif

  /*  Check if writing the input interval record to the buffer is possible  */
    old_buf_Nbytes_in_file = SLOG_typesz[ FrameHdr ]
                           + ( slog->inc_bbuf )->Nbytes_in_file
                           + ( slog->cur_bbuf )->Nbytes_in_file
                           + ( slog->out_bbuf )->Nbytes_in_file;
    new_buf_Nbytes_in_file = old_buf_Nbytes_in_file
                           + irec->bytesize;
#if defined( DEBUG )
    fprintf( stdout, __FILE__":SLOG_Irec_ToOutputStream() - \n" );
    fprintf( stdout, "       ""FrameHdr + inc_bbuf + cur_bbuf + out_bbuf = "
                     fmt_ui32" bytes\n", old_buf_Nbytes_in_file );
    fflush( stdout );
#endif

  /*
      Check if new buffer size is too big for the buffer
      If so, dump the content in the Bi-dir linked list buffer first
      before adding another interval record to the buffer
  */
    if ( new_buf_Nbytes_in_file > fbuf_bufsz( slog->fbuf ) ) {

        ierr = SLOG_STM_UpdateFRAME( slog );
        if ( ierr != SLOG_SUCCESS ) {
            fprintf( errfile, __FILE__":SLOG_Irec_ToOutputStream() - "
                              "SLOG_STM_UpdateFRAME( slog ) fails!!\n" );
            fflush( errfile );
            return SLOG_FAIL;
        }

        ierr = SLOG_STM_WriteFRAME( slog );
        if ( ierr == SLOG_FAIL ) {
            fprintf( errfile, __FILE__":SLOG_Irec_ToOutputStream() - "
                              "SLOG_STM_WriteFRAME( slog ) fails!!\n" );
            fflush( errfile );
            return SLOG_FAIL;
        }
        Nbytes_written = ierr;

        ierr = SLOG_STM_UpdateWriteFRAMEDIR( slog );
        if ( ierr != SLOG_SUCCESS ) {
            fprintf( errfile, __FILE__":SLOG_Irec_ToOutputStream() - "
                            "SLOG_STM_UpdateWriteFRAMEDIR( slog ) fails!!\n" );
            fflush( errfile );
            return SLOG_FAIL;
        }

        SLOG_STM_InitAgainBeforeAddRec( slog );

#if defined( DEBUG )
        /*  Check Number of Bytes written is consistent  */
        if ( Nbytes_written != old_buf_Nbytes_in_file ) {
            fprintf( errfile, __FILE__":SLOG_Irec_ToOutputStream() - "
                              "WARNING !!\n" );
            fprintf( errfile, "       ""SLOG_STM_WriteFRAME()'s "
                              "Nbytes_written {"fmt_ui32"} != SizeOf( "
                              "FrameHdr + inc_bbuf + cur_bbuf + out_bbuf "
                              ") {"fmt_ui32"}\n",
                              Nbytes_written , old_buf_Nbytes_in_file );
            fprintf( errfile, "        ""Continue with the program." );
            fflush( errfile );
        }
#endif
    
    }  /*  Endof { if ( new_buf_Nbytes_in_file > fbuf( slog->fbuf ) ) }  */

    /*
        since the new buffer size is NOT too big for the buffer 
        Add the new interval record to the buffer 
    */
    ierr = SLOG_Bbuf_AddIntvlRec( slog->cur_bbuf, irec );
    if ( ierr != SLOG_SUCCESS ) {
        fprintf( errfile, __FILE__":SLOG_Irec_ToOutputStream() - "
                          "SLOG_Bbuf_AddIntvlRec() fails !!\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    return SLOG_SUCCESS;
}



/*@C
    SLOG_CloseOutputStream - Close the specified SLOG_STREAM. It frees 
                             all the internally allocated memory used 
                             in SLOG_STREAM

  Modified Input Variables :
. slog - pointer to the SLOG_STREAM to be closed.

  Modifies Output Variables :
. returned value - integer status code

  Include File Needed :
  slog.h

.N SLOG_RETURN_STATUS
@*/
int SLOG_CloseOutputStream( SLOG_STREAM *slog )
{
    SLOG_uint32 Nbytes_written;
    SLOG_uint32 count_irec_in_bbufs;
    int         ierr;

    if ( slog->flagfirstentry == 0 ) {
#if defined( DEBUG )
        fprintf( stdout, __FILE__":SLOG_CloseOutputStream() - "
                 "Start SLOG_STM_UpdateFRAME()\n" );
        fprintf( stdout, __FILE__":SLOG_CloseOutputStream() - "
                 "FrameHdr+inc_bbuf+cur_bbuf+out_bbuf = %u bytes\n",
                 SLOG_typesz[ FrameHdr ]
                 + ( slog->inc_bbuf )->Nbytes_in_file
                 + ( slog->cur_bbuf )->Nbytes_in_file
                 + ( slog->out_bbuf )->Nbytes_in_file );
        fflush( stdout );
#endif

        count_irec_in_bbufs = ( slog->inc_bbuf )->count_irec
                            + ( slog->cur_bbuf )->count_irec
                            + ( slog->out_bbuf )->count_irec;

        if ( count_irec_in_bbufs > 0 ) { 
            ierr = SLOG_STM_UpdateFRAME( slog );
            if ( ierr != SLOG_SUCCESS ) {
                fprintf( errfile, __FILE__":SLOG_CloseOutputStream() - "
                                  "SLOG_STM_UpdateFRAME( slog ) fails!!\n" );
                fflush( errfile );
                return SLOG_FAIL;
            }

            ierr = SLOG_STM_WriteFRAME( slog );
            if ( ierr == SLOG_FAIL ) {
                fprintf( errfile, __FILE__":SLOG_CloseOutputStream() - "
                                  "SLOG_STM_WriteFRAME( slog ) fails!!\n" );
                fflush( errfile );
                return SLOG_FAIL;
            }
            Nbytes_written = ierr;
        
            ierr = SLOG_STM_UpdateWriteFRAMEDIR( slog );
            if ( ierr != SLOG_SUCCESS ) {
                fprintf( errfile, __FILE__":SLOG_CloseOutputStream() - "
                         "SLOG_STM_UpdateWriteFRAMEDIR( slog ) fails!!\n" );
                fflush( errfile );
                return SLOG_FAIL;
            }

            /* SLOG_STM_InitAgainBeforeAddRec( slog ); */
        }

    }   /*  Endof  if ( slog->flagfirstentry == 0 )  */

    ierr = SLOG_CloseStream( slog );
    if ( ierr != SLOG_SUCCESS ) {
        fprintf( errfile, __FILE__":SLOG_CloseOutputStream() - "
                          "SLOG_Close( slog ) fails!!\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

#if defined( DEBUG )
fprintf( stdout, __FILE__":SLOG_CloseOutputStream() - End\n" );
fflush( stdout );
#endif

    return SLOG_SUCCESS;
}



/*@C
    SLOG_Irec_SetMinRec - Set the fields for a bare minimal Interval Record

  Modified Input Variables :
. intvlrec - pointer to the internal interval record to be modified.

  Unmodified Input Variables :
. rectype - interval record's record field which determines if the
            record is a FIXED record or VARIABLE record, this in turns
            will determine the number of association and the number of 
            sets of MPI call arguments. 
. intvltype - interval record's interval type field.
. bebit_0 - interval record's first bebit.
. bebit_1 - interval record's second bebit.
. starttime - interval record's start time.
. duration - interval record's life time.
. node_id - interval record's node ID.
. cpu_id - interval record's cpu ID.
. thread_id - interval record's thread ID.
. where - interval record's where integer pointer.


  Modified Output Variables :
. returned value - integer return status.

  Usage Notes on this subroutine :
    The subroutine initializes an allocated internal interval
    record data structure.   So the routine must be called before
    appending the record to the SLOG file.

  Include File Needed :
  slog.h

.N SLOG_RETURN_STATUS
@*/
int SLOG_Irec_SetMinRec(       SLOG_intvlrec_t  *intvlrec,
                         const SLOG_rectype_t    rectype,
                         const SLOG_intvltype_t  intvltype,
                         const SLOG_bebit_t      bebit_0,
                         const SLOG_bebit_t      bebit_1,
                         const SLOG_starttime_t  starttime,
                         const SLOG_duration_t   duration,
                         const SLOG_nodeID_t     node_id,
                         const SLOG_cpuID_t      cpu_id,
                         const SLOG_threadID_t   thread_id,
                         const SLOG_where_t      where )
{
    if ( intvlrec == NULL ) {
        fprintf( errfile, __FILE__":SLOG_Irec_SetMinRec() - "
                          "Destination interval record pointer is NULL/n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    /*  Assemble all the input arguments as a SLOG_intvlrec_t, intvlrec  */
    intvlrec->rectype   = rectype    ;
    intvlrec->intvltype = intvltype ;
    intvlrec->bebits[0] = bebit_0  ;
    intvlrec->bebits[1] = bebit_1  ;
    intvlrec->starttime = starttime;
    intvlrec->duration  = duration ;
    intvlrec->node_id   = node_id  ;
    intvlrec->cpu_id    = cpu_id  ;
    intvlrec->thread_id = thread_id;
    intvlrec->where     = where    ;

    intvlrec->bytesize  = SLOG_typesz[ min_IntvlRec ];

    return SLOG_SUCCESS;
}



/*@C
    SLOG_Irec_SetAssocs - Set the association array of the interval record.

  Modified Input Variables :
. intvlrec - pointer to the internal Interval record to be modified.

  Unmodified Input Variables :
. slog  - pointer to the SLOG_STREAM where the Record Definition Table is 
          defined. 
. N_assocs - Number of elements in the association array.  This number will
             be checked to determine if it coincides with what is in
             Record Definition Table
. assocs - association array, array of file pointers to the associated 
           interval records of the specified interval record.


  Modified Output Variables :
. returned value - integer return status.

  Usage Notes on this subroutine :
    The subroutine appends the specified Internal Interval Record with 
    the given association array.  This subroutine is meant to be called
    once for each interval record written to the slogfile.

  Include File Needed :
  slog.h

@*/
int SLOG_Irec_SetAssocs(       SLOG_intvlrec_t  *intvlrec,
                         const SLOG_STREAM      *slog,
                         const SLOG_N_assocs_t   N_assocs,
                         const SLOG_assoc_t     *assocs  )
{
    SLOG_N_assocs_t recdefs_Nassocs;
    int             sizeof_assocs;
    int             ii;

    if ( intvlrec->bytesize == 0 ) {
        fprintf( errfile, __FILE__":SLOG_Irec_SetAssocs() - "
                          "SLOG_Irec_SetMinRec() needs to be called first\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    if ( ! SLOG_RecDef_IsVarIntvl( intvlrec->rectype ) ) {
        recdefs_Nassocs = SLOG_RecDef_NumOfAssocs( slog->rec_defs,
                                                   intvlrec->intvltype,
                                                   intvlrec->bebits[0],
                                                   intvlrec->bebits[1] );
        if ( N_assocs != recdefs_Nassocs ) {
            fprintf( errfile, __FILE__":SLOG_Irec_SetAssocs() - "
                             "Input Number of association does NOT match "
                             "the entry in Record Definition Table\n" );
            fprintf( errfile, "       ""The Input Number of associations = "
                              fmt_Nassocs_t"\n", N_assocs );
            fprintf( errfile, "       ""SLOG_RecDef_NumOfAssocs() returns "
                              fmt_Nassocs_t"\n", recdefs_Nassocs );
            fflush( errfile );
            return SLOG_FAIL;
        }
    }

  /*
      Read in the MPI call's corresponding association list 
      as determined by input intvltype
  */
    intvlrec->N_assocs = N_assocs;
    sizeof_assocs = intvlrec->N_assocs * sizeof( SLOG_assoc_t );
    intvlrec->assocs = ( SLOG_assoc_t * ) malloc( sizeof_assocs );
    if ( intvlrec->N_assocs > 0 && intvlrec->assocs == NULL ) {
        fprintf( errfile, __FILE__":SLOG_Irec_SetAssocs() - "
                          "malloc() for intvlrec->assocs fails\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    for ( ii = 0; ii < N_assocs; ii++ ) 
        intvlrec->assocs[ ii ] = assocs[ ii ];

    intvlrec->bytesize += sizeof_assocs;
    
    return SLOG_SUCCESS;
}



/*@C
    SLOG_Irec_SetArgs - Set the argument part of the interval record.

  Modified Input Variables :
. intvlrec - pointer to the internal interval record to be modified.

  Unmodified Input Variables :
. slog  - pointer to the SLOG_STREAM where the Record Definition Table 
          is defined.
. N_args - Number of elements in the argument array.  This number will
           be checked to determine if it coincides with what is in
           Record Definition Table
. args - argument array, array of arguments of the specified interval record.


  Modified Output Variables :
. returned value - integer return status.

  Usage Notes on this subroutine :
    The subroutine appends the specified Internal Interval with argument
    array.  For Fixed interval record, the subroutine is meant to called
    once for each specified interval record.  For variable interval 
    record, the subroutine can be called as many times as needed.

  Include File Needed :
  slog.h

@*/
int SLOG_Irec_SetArgs(       SLOG_intvlrec_t  *intvlrec,
                       const SLOG_STREAM      *slog,
                       const SLOG_N_args_t     N_args,
                       const SLOG_arg_t       *args )
{
    SLOG_vtrarg_t *vtr_args;
    SLOG_N_args_t  recdefs_Nargs;

    if ( SLOG_Irec_IsEmpty( intvlrec ) ) {
        fprintf( errfile, __FILE__":SLOG_Irec_SetArgs() - "
                          "SLOG_Irec_SetMinRec() needs to be called first\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    if ( ! SLOG_RecDef_IsVarIntvl( intvlrec->rectype ) ) {
        recdefs_Nargs = SLOG_RecDef_NumOfArgs( slog->rec_defs,
                                               intvlrec->intvltype,
                                               intvlrec->bebits[0],
                                               intvlrec->bebits[1] );
        if ( N_args != recdefs_Nargs ) {
            fprintf( errfile, __FILE__":SLOG_Irec_SetArgs() - "
                              "Input Number of arguments does NOT match "
                              "the entry in Record Definition Table\n" );
            fprintf( errfile, "       ""The Input Number of arguments = "
                              fmt_Nargs_t"\n", N_args );
            fprintf( errfile, "       ""SLOG_RecDef_NumOfArgs() returns "
                              fmt_Nargs_t"\n", recdefs_Nargs );
            fflush( errfile );
            return SLOG_FAIL;
        }
    }

    vtr_args = SLOG_Varg_Create( N_args, args );
    if ( SLOG_Irec_AddVtrArgs( intvlrec, vtr_args ) != SLOG_SUCCESS ) {
        fprintf( errfile, __FILE__":SLOG_Irec_SetArgs() - "
                          "SLOG_Irec_AddVtrArgs() fails\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    if ( SLOG_RecDef_IsVarIntvl( intvlrec->rectype ) )
        intvlrec->bytesize += sizeof( SLOG_N_args_t );
    intvlrec->bytesize += N_args * sizeof( SLOG_arg_t );

    return SLOG_SUCCESS;
}
