#include <stdio.h>
#include "slog_recdefs.h"              /*I "slog_recdefs.h" I*/



void SLOG_RDEF_Assign(       SLOG_recdef_t        *recdef,
                       const SLOG_intvltype_t      in_intvltype,
                       const SLOG_bebit_t          in_bebit_0,
                       const SLOG_bebit_t          in_bebit_1,
                       const SLOG_N_assocs_t       in_Nassocs,
                       const SLOG_N_args_t         in_Nargs )
{
    recdef->intvltype = in_intvltype;
    recdef->bebits[0] = in_bebit_0;
    recdef->bebits[1] = in_bebit_1;
    recdef->Nassocs   = in_Nassocs;
    recdef->Nargs     = in_Nargs;
}



int SLOG_RDEF_Duplicated( const SLOG_recdefs_table_t  *recdefs,
                          const SLOG_recdef_t         *in_def )
{
    SLOG_recdef_t  *cur_def;
    int             ii;
    int             NotMatched = 1;

    for ( ii = 0; NotMatched && (ii < recdefs->Nentries); ii++ ) {
        cur_def = &( recdefs->entries[ ii ] );
        NotMatched = !(    ( cur_def->intvltype == in_def->intvltype )
                        && ( cur_def->bebits[0] == in_def->bebits[0] )
                        && ( cur_def->bebits[1] == in_def->bebits[1] ) ); 
    }

    return ! NotMatched;
}


/*@C
    SLOG_RDEF_Open - Creates the SLOG Record Definition Table, 
                     SLOG_recdefs_table_t, needed to write the to
                     the disk.

  Modified Input Variables :
. slog - pointer to the SLOG_STREAM where SLOG Record Definition Table is
         located.

  Modified Output Variables :
. returned value - integer return status.

  Include file needed : 
  slog_recdefs.h

.N SLOG_RETURN_STATUS
@*/
int SLOG_RDEF_Open( SLOG_STREAM  *slog )
{
    slog->rec_defs = ( SLOG_recdefs_table_t * )
                     malloc( sizeof( SLOG_recdefs_table_t ) );
    if ( slog->rec_defs == NULL ) {
        fprintf( errfile, __FILE__":SLOG_RDEF_Open() - "
                          "malloc() fails, Cannot allocate memory for "
                          "SLOG_recdefs_table_t\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    slog->rec_defs->increment  = 10;
    slog->rec_defs->capacity   = slog->rec_defs->increment;
    slog->rec_defs->Nentries   = 0;
    slog->rec_defs->entries    = ( SLOG_recdef_t * )
                                 malloc(   slog->rec_defs->capacity
                                         * sizeof( SLOG_recdef_t ) );

    return SLOG_SUCCESS;
}



/*@C
    SLOG_RDEF_AddRecDef - Add one definition to the SLOG Record Definition
                          table.

  Modified Input Variables :
. slog - pointer to the SLOG_STREAM where SLOG Record Definition Table is
         located.

  Unmodified Input Variables :
. intvltype - label to the interval type.
. bebit_0 - the 1st bebit.
. bebit_1 - the 2nd bebit.
. Nassocs - Number of assocications for this interval record type.
. Nargs - Number of MPI call argument for this interval record type.

  Modified Output Variables :
. returned value - integer return status.

  Include file needed : 
  slog_recdefs.h

.N SLOG_RETURN_STATUS
@*/
int SLOG_RDEF_AddRecDef(       SLOG_STREAM          *slog,
                         const SLOG_intvltype_t      intvltype,
                         const SLOG_bebit_t          bebit_0,
                         const SLOG_bebit_t          bebit_1,
                         const SLOG_N_assocs_t       Nassocs,
                         const SLOG_N_args_t         Nargs )
{
    SLOG_uint32             idx;
    SLOG_recdefs_table_t   *recdefs;

    recdefs = slog->rec_defs;

#if defined( DEBUG )
    if ( recdefs == NULL ) {
        fprintf( errfile, __FILE__":SLOG_RDEF_AddRecDef() - "
                          "the SLOG_recdefs_table_t pointer in slog "
                          "is NULL\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }
#endif

    /*  Check if there is any space left for a new record definition */
    if ( recdefs->Nentries + 1 >= recdefs->capacity ) {
        recdefs->capacity += recdefs->increment;
        recdefs->entries = ( SLOG_recdef_t * )
                           realloc( recdefs->entries,
                                    recdefs->capacity * sizeof( SLOG_recdef_t )
                                  );
        if ( recdefs->entries == NULL ) {
            fprintf( errfile, __FILE__":SLOG_RDEF_AddRecDef() - "
                              "realloc fails, cannot increase the size "
                              "of recdefs->entries[]\n" );
            fprintf( errfile, "\t""The input record definition = [ "fmt_itype_t
                              ", ("fmt_bebit_t", "fmt_bebit_t
                              "), "fmt_Nassocs_t", "fmt_Nargs_t" ] \n",
                              intvltype, bebit_0, bebit_1, Nassocs, Nargs );
            fflush( errfile );
            return SLOG_FAIL;
        }
    }

    
    idx = recdefs->Nentries;
    SLOG_RDEF_Assign( &( recdefs->entries[ idx ] ),
                      intvltype, bebit_0, bebit_1, Nassocs, Nargs );
    if ( SLOG_RDEF_Duplicated( recdefs, &(recdefs->entries[ idx ]) ) ) {
        fprintf( errfile, __FILE__":SLOG_RDEF_AddRecDef() - "
                          "there is a copy of input record\n " );
        fprintf( errfile, "\t""definition with identical keys "
                          "in the SLOG record definition table\n" );
        fprintf( errfile, "\t""The input record definition = [ "fmt_itype_t
                          ", ("fmt_bebit_t", "fmt_bebit_t
                          "), "fmt_Nassocs_t", "fmt_Nargs_t" ] \n",
                          intvltype, bebit_0, bebit_1, Nassocs, Nargs );
        fflush( errfile );
        return SLOG_FAIL;
    }
    recdefs->Nentries += 1;

    return SLOG_SUCCESS;
}


/*@C
    SLOG_RDEF_Close - Write the SLOG_recdefs_table_t onto the disk

  Modified Input Variables :
. slog - pointer to the SLOG_STREAM where SLOG Record Definition Table is
         located.

  Modifies Output Variables :
. returned value - integer status code.

  Include file needed : 
  slog_recdefs.h

.N SLOG_RETURN_STATUS
@*/
int SLOG_RDEF_Close( SLOG_STREAM  *slog )
{
    SLOG_recdefs_table_t   *recdefs;
    SLOG_recdef_t          *def;
    SLOG_bebits_t           bebits;
    int                     ierr;
    int                     ii;

    /*  Update the content of SLOG_hdr_t to both memory and disk  */
    slog->hdr->fptr2recdefs = ( SLOG_fptr ) ftell( slog->fd );
    ierr = SLOG_WriteUpdatedHeader( slog );
    if ( ierr != SLOG_SUCCESS ) {
        fprintf( errfile,  __FILE__":SLOG_RDEF_Close() - "
                                   "SLOG_WriteUpdatedHeader() fails\n" );
        fflush( errfile );
        return ierr;
    }

    /*  Write the content of SLOG Record Definition Table to the disk  */
    recdefs = slog->rec_defs;
    
    bswp_fwrite( &( recdefs->Nentries ),
                 sizeof( SLOG_uint32 ), 1, slog->fd );
    for ( ii = 0; ii < recdefs->Nentries; ii++ ) {
        def = &( recdefs->entries[ ii ] );
        bswp_fwrite( &( def->intvltype ),
                     sizeof( SLOG_intvltype_t ), 1, slog->fd );
        bebits = SLOG_bebits_encode( def->bebits[0], def->bebits[1] );
        bswp_fwrite( &( bebits ),
                     sizeof( SLOG_bebits_t ), 1, slog->fd );
        bswp_fwrite( &( def->Nassocs ),
                     sizeof( SLOG_N_assocs_t ), 1, slog->fd );
        bswp_fwrite( &( def->Nargs ),
                     sizeof( SLOG_N_args_t ), 1, slog->fd );
    }
  
    return SLOG_SUCCESS;
}



void SLOG_RDEF_Free( SLOG_recdefs_table_t *recdefs )
{
    if ( recdefs != NULL ) {
        if ( recdefs->Nentries > 0 && recdefs->entries != NULL )
            free( recdefs->entries );
        free( recdefs );
    }
}



int SLOG_RDEF_ReadRecDefs( SLOG_STREAM  *slog )
{
    SLOG_recdefs_table_t   *recdefs;
    SLOG_recdef_t          *def;
    SLOG_bebits_t           bebits;
    int                     ii;
    int                     ierr;

    slog->rec_defs = ( SLOG_recdefs_table_t * )
                     malloc( sizeof( SLOG_recdefs_table_t ) );
    if ( slog->rec_defs == NULL ) {
        fprintf( errfile, __FILE__":SLOG_RDEF_ReadRecDefs() - "
                          "malloc() fails, Cannot allocate memory for "
                          "SLOG_recdefs_table_t\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    /*  Initialize the Record Definition table data structure  */
    slog->rec_defs->increment  = 10;
    slog->rec_defs->capacity   = slog->rec_defs->increment;
    slog->rec_defs->Nentries   = 0;
    slog->rec_defs->entries    = NULL;
    
    /*  Seek to the beginning of the Record Def'n Table in the SLOG file  */
    if ( slog->hdr->fptr2recdefs == SLOG_fptr_NULL ) {
        fprintf( errfile, __FILE__":SLOG_RDEF_ReadRecDefs() - "
                          "slog->hdr->fptr2recdefs == SLOG_fptr_NULL\n" );
        fprintf( errfile, "\t""SLOG_RDEF_Close() may NOT be called\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }
    ierr = fseek( slog->fd, slog->hdr->fptr2recdefs, SEEK_SET );
    if ( ierr != 0 ) {
        fprintf( errfile, __FILE__":SLOG_RDEF_ReadRecDefs() - "
                          "fseek( fptr2recdefs ) fails\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    /*  Read in the data structure  */
    recdefs = slog->rec_defs;
    
    bswp_fread( &( recdefs->Nentries ), sizeof( SLOG_uint32 ), 1, slog->fd );
    recdefs->capacity = recdefs->Nentries;
    recdefs->entries  = ( SLOG_recdef_t * )
                        malloc( recdefs->capacity * sizeof( SLOG_recdef_t ) );
    if ( recdefs->Nentries > 0 && recdefs->entries == NULL ) {
        fprintf( errfile, __FILE__":SLOG_RDEF_ReadRecDefs() - "
                          "malloc() fails for recdefs->entries[]\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }
    
    for ( ii = 0; ii < recdefs->Nentries; ii++ ) {
        def = &( recdefs->entries[ ii ] );
        bswp_fread( &( def->intvltype ),
                    sizeof( SLOG_intvltype_t ), 1, slog->fd );
        bswp_fread( &( bebits ),
                    sizeof( SLOG_bebits_t ), 1, slog->fd );
        SLOG_bebits_decode( bebits, def->bebits );
        bswp_fread( &( def->Nassocs ),
                    sizeof( SLOG_N_assocs_t ), 1, slog->fd );
        bswp_fread( &( def->Nargs ),
                    sizeof( SLOG_N_args_t ), 1, slog->fd );
    }
  
    return SLOG_SUCCESS;
}


void SLOG_RDEF_Print( FILE* fd, const SLOG_recdefs_table_t *recdefs )
{
    SLOG_recdef_t          *def;
    int ii;

    for ( ii = 0; ii < recdefs->Nentries; ii++ ) {
        def = &( recdefs->entries[ ii ] );
        fprintf( fd, " def[%i] = [ "fmt_itype_t", ("fmt_bebit_t", "
                     fmt_bebit_t"), "fmt_Nassocs_t", "fmt_Nargs_t" ]\n", ii,
                     def->intvltype, def->bebits[0], def->bebits[1],
                     def->Nassocs, def->Nargs );
        fflush( fd );
    }
}
